import sys
import logging
import os
from logging import getLogger


def exit_with_error(message, exit_status, logger, tmp_folder=None):
    """
    Print an error message to stderr, prefixed by the program name and 'ERROR'.
    Then exit program with supplied exit status.
    :param message: Message to give the user upon exit
    :param exit_status: Status returned as exit status
    :param tmp_folder: Temporary folder for Corekaburra to be deleted under some circumstances.
    :param logger: Logger for program
    :return: None
    """

    # Delete tmp files and folder
    try:
        if tmp_folder is not None:
            tmp_files = os.listdir(tmp_folder)
            for file in tmp_files:
                os.remove(os.path.join(tmp_folder, file))
            os.rmdir(tmp_folder)
        else:
            pass
    except FileNotFoundError:
        pass

    logger.error(message)
    print(f"Corekaburra ERROR: {message}, exiting", file=sys.stderr)
    sys.exit(exit_status)



import os

try:
    from Corekaburra.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error
EXIT_INPUT_FILE_ERROR = 1


def parse_complete_genome_file(complete_genome_file, gff_files, logger):
    """
    Function to check if all genomes given as complete genomes can be found in the pan genome.
    :param complete_genome_file:
    :param gff_files:
    :param logger: Logger for program
    :return: a list of the base name of the complete genomes.
    """

    # Read the file and all lines (complete genomes given)
    with open(complete_genome_file, 'r') as genome_file:
        complete_genomes = genome_file.readlines()
        complete_genomes = [name.strip().replace('.gz', '') for name in complete_genomes]
        complete_genomes = [name.replace('.gff', '') for name in complete_genomes]
        complete_genomes = [os.path.basename(name) for name in complete_genomes]

    # Take input gffs and remove path to the file
    gffs = [os.path.basename(gff).replace('.gff', '').replace('.gz', '') for gff in gff_files]

    # check that all complete genomes are in the input gffs
    complete_genome_status = all(complete_genome in gffs for complete_genome in complete_genomes)

    # If the complete genomes are found, return a list of complete genomes
    if complete_genome_status:
        logger.debug(f'complete genomes: {complete_genomes = } were identified and accepted')
        return complete_genomes
    else:
        logger.debug(f'complete genomes: {complete_genomes = } were identified but not accepted!')
        exit_with_error('Genome given in Complete genomes was not identified in pan-genome!', EXIT_INPUT_FILE_ERROR, logger)


if __name__ == '__main__':
    pass



import csv
import os
import time


def master_info_writer(master_info, out_path, prefix):
    """
    Function to write two output .tsv files related to regions content and size for each genome
    :param master_info: Dict of info for each core gene pair across all genomes
    :param out_path: Path to the output folder
    :param prefix: A possible prefix for the output files.
    :return: Nothing
    """

    # Write general content
    out_file_name = 'low_frequency_gene_placement.tsv' # Previously 'low_frequency_gene_placement.tsv' - Proposed name: core_region_content.tsv
    if prefix is not None:
        out_file_name = f'{prefix}_{out_file_name}'
    with open(os.path.join(out_path, out_file_name), 'w', newline='', encoding='utf-8') as out_file:
        writer = csv.writer(out_file, delimiter="\t")

        # Create header
        header = ['Gff', 'Core_gene_1', 'Core_gene_2', 'Core_region_size',
                  'Core_region_accessory_count']
        writer.writerow(header)

        # Write remaining rows:
        for key in sorted(master_info.keys()):
            info = master_info[key][0:5]

            writer.writerow(info)

    # Write gene content in long format
    out_file_name = 'core_core_accessory_gene_content.tsv' # Previously core_core_accessory_gene_content.tsv - Proposed name: accessory_gene_placement.tsv
    if prefix is not None:
        out_file_name = f'{prefix}_{out_file_name}'

    with open(os.path.join(out_path, out_file_name), 'w', newline='', encoding='utf-8') as out_file:
        writer = csv.writer(out_file, delimiter="\t")

        # Create header
        header = ['Gff', 'Core_gene_1', 'Core_gene_2', 'gene', 'type']
        writer.writerow(header)

        # Write remaining rows:
        for key in sorted(master_info.keys()):
            core_core_region = master_info[key]
            if len(core_core_region[5]):
                for gene in core_core_region[5]:
                    row = [core_core_region[0],
                           core_core_region[1],
                           core_core_region[2],
                           gene,
                           'intermediate_frequency']
                    writer.writerow(row)

            if len(core_core_region[6]):
                for gene in core_core_region[6]:
                    row = [core_core_region[0],
                           core_core_region[1],
                           core_core_region[2],
                           gene,
                           'low_frequency']
                    writer.writerow(row)


def summary_info_writer(master_summary_info, out_path, prefix):
    """
    Function for writing the summary table for regions identified across genomes
    :param master_summary_info: Dict holding summary statistics for core pair region identified
    :param out_path: Path to the output folder
    :param prefix: Prefix for any output files
    :return: Nothing
    """
    # Generate file name
    out_file_name = 'core_pair_summary.csv' # Previously: core_pair_summary.csv - proposed name: core_region_summary.csv
    if prefix is not None:
        out_file_name = prefix + '_' + out_file_name

    # Write general content
    with open(os.path.join(out_path, out_file_name), 'w', newline='', encoding='utf-8') as out_file:
        writer = csv.writer(out_file)

        # Create header
        header = ['Core_pair', 'n',
                  'occurrence_core_1', 'occurrence_core_2', 'co_occurrence',
                  'min_dist', 'max_dist', 'mean_dist', 'median_dist',
                  'min_acc', 'max_acc', 'mean_acc', 'median_acc']
        writer.writerow(header)

        # Write remaining rows:
        for key in sorted(master_summary_info.keys()):
            info = master_summary_info[key]

            writer.writerow(info)


def segment_writer(segments, out_path, prefix):
    """
    Function to write segments of core genes identified across the pan-genome
    :param segments: Dict of segments (lists) in values, under name of segments as keys.
    :param out_path: Path to output folder
    :param prefix: Prefix for any output files
    :return: Nothing
    """

    # Generate file name
    out_file_name = 'core_segments.csv'
    if prefix is not None:
        out_file_name = prefix + '_' + out_file_name

    # Write general content
    with open(os.path.join(out_path, out_file_name), 'w', newline='', encoding='utf-8') as out_file:
        writer = csv.writer(out_file)

        # Create header
        header = ['Segment_name', 'Segment_position', 'Core_gene']
        writer.writerow(header)

        # Write remaining rows:
        for key in sorted(segments.keys()):
            for index, gene in enumerate(segments[key]):
                info = [key.replace('--', '-'), index+1, gene]

                writer.writerow(info)


def no_acc_segment_writer(no_acc_segments, out_path, prefix):
    """
    Function for writing segments of core genes with no accessory between them.
    :param no_acc_segments: Dict of segments with (lists) in values with sub-lists being segments with no accessory genes between them, under name of segments as keys.
    :param out_path: Path to output folder
    :param prefix: Prefix for any output files
    :return: Nothing
    """

    # Generate file name
    out_file_name = 'no_accessory_core_segments.csv'
    if prefix is not None:
        out_file_name = prefix + '_' + out_file_name

    # Write general content
    with open(os.path.join(out_path, out_file_name), 'w', newline='', encoding='utf-8') as out_file:
        writer = csv.writer(out_file)

        # Create header
        header = ['Parent_segment_name', 'Sub_segment_name', 'Parent_segment_position', 'Sub_segment_position', 'Core_gene']
        writer.writerow(header)

        # Write remaining rows:
        for key in sorted(no_acc_segments.keys()):
            for sub_index, subsegment in enumerate(no_acc_segments[key]):
                sub_name = f'{subsegment[0]}-{subsegment[-1]}'
                for index, gene in enumerate(subsegment):
                    info = [key.replace('--', '-'), sub_name, sub_index + 1, index + 1, gene]

                    writer.writerow(info)


def non_core_contig_writer(non_core_contigs, out_path, prefix,):
    """
    Function to write output for contigs with no core gene, but with accessory genes
    :param non_core_contigs: Dict of info for each contig with no core genes, values are list of lists with intermediate and low-frequency genes
    :param out_path: Path to the output folder
    :param prefix: A possible prefix for the output files.
    :return: Nothing
    """
    # Write gene content in long format
    out_file_name = 'coreless_contig_accessory_gene_content.tsv'
    if prefix is not None:
        out_file_name = f'{prefix}_{out_file_name}'

    with open(os.path.join(out_path, out_file_name), 'w', newline='', encoding='utf-8') as out_file:
        writer = csv.writer(out_file, delimiter="\t")

        # Create header
        header = ['Gff', 'Contig', 'Accessory_count', 'Intermediate_cunt', 'low_frequency_count']
        writer.writerow(header)

        # Write remaining rows:
        for key in sorted(non_core_contigs):
            genome, contig = key.split('--')
            num_intermidiate = len(non_core_contigs[key][0])
            num_low = len(non_core_contigs[key][1])
            num_accessory = num_intermidiate + num_low

            info = [genome, contig, num_accessory, num_intermidiate, num_low]

            writer.writerow(info)


if __name__ == "__main__":
    pass



import numpy as np

try:
    from Corekaburra.consesus_core_genome import count_gene_co_occurrence
except ModuleNotFoundError:
    from consesus_core_genome import count_gene_co_occurrence


def calculate_n_create_summaries(master_info, core_gene_dict):
    """
    Function used to calculate summary statistics for core gene pairs
    :param master_info: Dict over master info from core pairs identified from gff files
    :param core_gene_dict: Dict mapping genes to genomes and pan-genome cluster (genes)
    :return: Dict containing the summary statistics for each core pair.
    """
    # initialize the summary dict to be returned
    summary_dict = {}

    # Group the values of length of accessory gene content for the core pairs into the summary dict
    for pair in master_info:
        core_pair = pair.rsplit('--', 1)[0]

        # Try to add the values for a given core pair to the summary dict,
        # if not possible then add a two column array to they key and add the values
        try:
            summary_dict[core_pair].append(master_info[pair][3:5])
        except KeyError:
            summary_dict[core_pair] = [master_info[pair][3:5]]

    occurrence_dict = {key: {} for key in summary_dict}
    # dict.fromkeys(summary_dict.keys(), {})
    for core_pair in summary_dict:
        gene_list = core_pair.split('--')
        # Calculate the occurrence and co-occurrence of core genes

        # Calculate gene occurrence and co-occurrence
        occurrence_dict[core_pair]['co_occurrence'], individual_occurrences = count_gene_co_occurrence(core_gene_dict, gene_list)

        for gene in gene_list:
            occurrence_dict[core_pair][gene] = individual_occurrences[gene]

    # Take the mean over length of accessory genes for all core-pairs
    for core_pair in summary_dict:
        gene_list = core_pair.split('--')
        # cast the list into a numpy array, format it to have lengths and accessory gene count as a column each
        summary_dict[core_pair] = np.array(summary_dict[core_pair])
        summary_dict[core_pair] = summary_dict[core_pair].reshape((-1, 2))

        # Calculate summary statistics
        pair_occurrence = summary_dict[core_pair].shape[0]
        min_values = np.min(summary_dict[core_pair], axis=0)
        max_values = np.max(summary_dict[core_pair], axis=0)
        mean_values = np.mean(summary_dict[core_pair], axis=0)
        median_values = np.median(summary_dict[core_pair], axis=0)

        # Add the summary to the dict, round mean and median to one decimal
        summary_dict[core_pair] = [core_pair.replace('--', '-'),
                                   pair_occurrence,
                                   occurrence_dict[core_pair][gene_list[0]],
                                   occurrence_dict[core_pair][gene_list[1]],
                                   occurrence_dict[core_pair]['co_occurrence'], # TODO - Add neighbour ratio?
                                   min_values[0], max_values[0],
                                   round(mean_values[0], 1), round(median_values[0], 1),
                                   min_values[1], max_values[1],
                                   round(mean_values[1], 1), round(median_values[1], 1)
                                   ] # TODO - at the moment all sequence breaks are reported as zero and gives a co-occurrence equal to zero - is this acceptable?

    return summary_dict


if __name__ == '__main__':
    pass



def merge_dicts_counts(parent_dict, merge_object):
    """
    Function that can merge two dicts by keys and adding 1 to the value each time key is observed
    :param parent_dict: Dict to which the second with should be merged into
    :param merge_object: Dict or List to be merged into the first.

    :return: Resulting dict following merge
    """

    if isinstance(merge_object, dict):
        keys = merge_object.keys()
    elif isinstance(merge_object, list):
        keys = merge_object

    for key in keys:
        if key in parent_dict:
            parent_dict[key] += 1
        else:
            parent_dict[key] = 1

    return parent_dict


def merge_dicts_lists(parent_dict, merge_dict):
    """
    Function to add two dictionaries by adding lists of matching keys
    :param parent_dict: The Dict to which the second dict should be merged with
    :param merge_dict: Dict to be merge with the parent

    :return: Dict having the two inputs merged
    """

    for key in merge_dict.keys():
        # Check if key is present, if then append the value to the key
        if key in parent_dict:
            if not isinstance(merge_dict[key], list):
                parent_dict[key] += [merge_dict[key]]
            else:
                parent_dict[key] += merge_dict[key]

        # If key is not present construct the key
        else:
            if not isinstance(merge_dict[key], list):
                parent_dict[key] = [merge_dict[key]]
            else:
                parent_dict[key] = merge_dict[key]

    return parent_dict



import os
import warnings

try:
    from Corekaburra.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error
EXIT_INPUT_FILE_ERROR = 1
EXIT_COMMAND_LINE_ERROR = 2


def check_cutoffs(low_cutoff, core_cutoff, logger):
    """
    Function to check the given cutoffs are legal, otherwise provide more info.
    :param low_cutoff: Cutoff for low-frequency genes
    :param core_cutoff: Cutoff for core genes
    :param logger: Program logger
    :return: Nothing
    """
    if 0 <= low_cutoff < core_cutoff <= 1:
        logger.debug(f'User provided cutoffs: {low_cutoff = } and {core_cutoff = } were accepted')
        return
    else:
        exit_with_error('Something is wrong with cutoffs for core and low-frequency genes!\n'
                        'Make sure the cutoff for core genes is larger than for low-frequency, and is >0 or =1.\n'
                        'Also make sure that the low-frequency gene cutoff is either equal to 0 or <1',
                        EXIT_COMMAND_LINE_ERROR, logger)


def define_pangenome_program(folder, logger):
    """
    Function to examine if input pan genome folder stems from Roary or Panaroo.
    :param folder: Input folder provided as pan-genome folder
    :param logger: Program logger
    :return: The name of the program from which the pangenome is suspected to come from
    """

    logger.debug("----Checking presence of input files in pan genome folder----")

    try:
        if os.path.isfile(os.path.join(folder, 'gene_presence_absence.csv')):
            # See if input is from Roary
            with open(os.path.join(folder, 'gene_presence_absence.csv'), 'r') as gene_pres_abs:
                if '"' in gene_pres_abs.readline():
                    gene_pres_abs_file_path = os.path.join(folder, 'gene_presence_absence.csv')
                    logger.debug('Pan-genome_program detected to be: Roary')
                    return "Roary", gene_pres_abs_file_path

        # See if input is from Panaroo
        gene_pres_abs_file_path = os.path.join(folder, 'gene_presence_absence_roary.csv')
        if os.path.isfile(gene_pres_abs_file_path):
            logger.debug('Pan-genome_program detected to be: Panaroo')
            return "Panaroo", gene_pres_abs_file_path
        else:
            exit_with_error('No gene presence/absence file was found in given pan-genome folder', EXIT_INPUT_FILE_ERROR, logger)

    except FileNotFoundError:
        exit_with_error('No gene presence/absence file was found in given pan-genome folder', EXIT_INPUT_FILE_ERROR)


def check_gene_data(folder, logger):
    """
    Check if the gene_data.csv file is present in the folder from a Panaroo pan-genome run
    :param folder: Input folder provided as pan-genome folder
    :param logger: Program logger
    :return: Path to the identified gene_data.csv file
    """

    logger.debug('Identify gene_data.csv from Panaroo')

    if os.path.isfile(os.path.join(folder, 'gene_data.csv')):
        logger.debug('Gene_data.csv from Panaroo found!')
        return os.path.join(folder, 'gene_data.csv')
    else:
        exit_with_error('gene_data.csv file could not be located in the given pan genome input folder.\n'
                        'Please give the -a flag to omit this step or locate the gene_data.csv file.',
                        EXIT_INPUT_FILE_ERROR, logger)


def check_gff_in_pan(file_list, gene_presence_absence_path, logger):
    """
    Function to check if all given Gff files are in the given pan-genome
    :param file_list: List of Gff file paths
    :param gene_presence_absence_path: File path to the identified gene_presence_absence_file
    :param logger: Program logger
    :return: Bool used for unit testing
    """
    with open(gene_presence_absence_path, 'r') as pan_file:
        # Read the first line of the gene_presence_absence and extract the genome names
        pan_header_line = pan_file.readline()
        pan_header_line = pan_header_line.strip().split(',')
        genome_names = pan_header_line[14:]
        genome_names = [name.replace('"', '') for name in genome_names]

    file_list = [os.path.basename(file) for file in file_list]
    file_list_no_suffix = [file.rstrip('.gff') for file in file_list]

    # Check if all or subset of GFFs from pan genome have been supplied,
    # if only a subset then raise warning
    if set(file_list).issubset(genome_names) or set(file_list_no_suffix).issubset(genome_names):
        if len(file_list) < len(genome_names):
            logger.info("\nNot all gff in pan genome given as input. I will run with it but are you sure this is deliberate?\n")

        return True  # True used for unit testing

    # Exit with error is not all inputs can be found in the pan-genome presence absence file
    exit_with_error('Unexpected occurrence in the matching of input GFF files and the pan genome presence/absence file',
                    EXIT_INPUT_FILE_ERROR, logger)



import networkx as nx
try:
    from Corekaburra.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error
EXIT_SEGMENT_IDENTIFICATION_ERROR = 4

# pylint: disable=E1123, E1121


def construct_core_graph(core_neighbour_pairs):
    """
    Function to construct a graph from the core pairs and number of times each is observed.
    :param core_neighbour_pairs: Dict of core pairs and the number of times each is observed
    :return: A graph with nodes being core genes, edges being a connection between them with a weight of the number of times they are connected
    """
    # Initiate core gene graph
    G = nx.Graph()

    # Add all core gene pairs and their edges
    for core_set in core_neighbour_pairs.keys():
        # split core genes
        core_genes = core_set.split('--')

        # Check that sequence break is not present:
        if 'Sequence_break' != core_genes[0] and 'Sequence_break' != core_genes[1]:
            # Construct edge list
            core_genes = [(core_genes[0], core_genes[1], {'weight': core_neighbour_pairs[core_set]})]
            # Add edge
            G.add_edges_from(ebunch_to_add=core_genes)

    return G


def count_gene_co_occurrence(core_gene_dict, two_gene_segment):
    """
    Function to find the number of genomes in which two genes co-occur across the input genomes.
    :param core_gene_dict: Dictionary over core genes mapped from genome, to locus_tag, to pan-genome cluster
    :param two_gene_segment: List of two genes forming a segment
    :return: Int - number of co-occurrences for the two genes in the input two_gene_segment
    """
    co_occurrence = 0
    gene_occurrence = dict.fromkeys(two_gene_segment, 0)

    # Get pan-genome clusters for all genomes in a list of lists
    core_gene_presences = [list(core_genes.values()) for core_genes in core_gene_dict.values()]

    # Go through all genomes and check if genes co-occur
    for core_gene_set in core_gene_presences:
        # count the co-occurrences
        if set(two_gene_segment).issubset(core_gene_set):
            co_occurrence += 1

        # Count the individual occurrences
        if two_gene_segment[0] in core_gene_set:
            gene_occurrence[two_gene_segment[0]] += 1
        if two_gene_segment[1] in core_gene_set:
            gene_occurrence[two_gene_segment[1]] += 1

    return co_occurrence, gene_occurrence


def identify_no_accessory_segments(double_edge_segements, combined_acc_gene_count):
    """
    Function that takes segments between multi connected core genes, and a Dict of number of accessory genes in core-core regions.
    Divides segments into smaller subsegments, in which no accessory genes can be found between core pairs.
    :param double_edge_segements: Dict of segments of core genes identified. Keys are genes at edges of segments. Value is a List of genes in the segment from one side to the other.
    :param combined_acc_gene_count: Dict of the number of accessory genes (value) identified between a set core genes (Key)
    :return: Dict of subsegments. Same keys as for the segment dict, but keys are a list of lists. Each sub-list is a subsegment.
    """
    # TODO - ATM this does not occur if every core gene only has two connections. Should is still occur to let a complete static genome synteny be divided into no-accessory segments?
    # Create dict of subsegments of the larger segments
    sub_segment_dict = {key: [] for key in double_edge_segements}

    # Go through segments to identify subsegments
    for segment in double_edge_segements:
        empty_segment_genes = []

        cur_segment = double_edge_segements[segment]
        # Check each region of the segment for core genes
        for i in range(0, len(cur_segment)-1):
            core_neighbours = sorted([cur_segment[i], cur_segment[i+1]])
            core_region = f'{core_neighbours[0]}--{core_neighbours[1]}'
            # Get accessory genes in region
            core_region_acc_genes = combined_acc_gene_count[core_region]

            # If core region does not contain accessory genes, add to current segment.
            # Else add the segment and start a new
            if core_region_acc_genes == 0:
                # If  first pair in segment add both, if not first only add the last gene
                if len(empty_segment_genes) == 0:
                    empty_segment_genes += [cur_segment[i], cur_segment[i+1]]
                else:
                    empty_segment_genes += [cur_segment[i+1]]

            else:
                # Check if first pair in subsegment and add first gene as being 'lonely'
                if len(empty_segment_genes) == 0:
                    empty_segment_genes += [cur_segment[i]]

                # Record the segment and reset the subsegment to contain no core genes
                sub_segment_dict[segment].append(empty_segment_genes)
                empty_segment_genes = []

            # Check of segment end has been reached and more than two genes are in the segment, if then add the segment
            if i == len(cur_segment) - 2 and len(empty_segment_genes) >= 2:
                sub_segment_dict[segment].append(empty_segment_genes)
                empty_segment_genes = []
            # Check if the second gene in pair is last in segment, and accessory genes are present between second to last and last core gene,
            # if then add the last gene as being 'lonely'
            elif i == len(cur_segment) - 2:
                empty_segment_genes += [cur_segment[i + 1]]
                sub_segment_dict[segment].append(empty_segment_genes)

    return sub_segment_dict


def identify_segments(core_graph, num_gffs, core_gene_dict):
    """
    Function to identify stretches of core genes between core genes neighbouring multiple different genes
    :param core_graph: Graph over core genes with weights being the number of connections between the genes
    :param num_gffs: Number of gffs inputted
    :return: Dict over stretches of core genes found in the core gene graph.
    """

    # Identify all nodes that contain more than two degrees.
    multi_edge_nodes = [node for node, connections in core_graph.degree if connections > 2]

    # Check if any node have multiple edges, if not then return.
    if len(multi_edge_nodes) == 0:
        return None

    # Dict to hold connections between >2 edge nodes
    connect_dict = {}

    # for all nodes with >2 degrees themself, identify neighbouring nodes with >2 degrees
    for node in multi_edge_nodes:
        connect_dict[node] = [neighbor for neighbor in core_graph.neighbors(node) if neighbor in multi_edge_nodes]

    # Turn the weight into a 'distance' or number of times not found together.
    for edge in core_graph.edges(data=True):
        core_graph[edge[0]][edge[1]]['weight'] = num_gffs - core_graph[edge[0]][edge[1]]['weight']

    # find all simple paths between nodes with >2 degrees
    double_edge_segements = {}
    multi_edge_connect_adjust = []

    # Go through all source and taget nodes,
    # see if a path can be found where all nodes between them have only two degrees
    for source_node in multi_edge_nodes:
        for target_node in multi_edge_nodes:
            if target_node != source_node:
                # Get path (segment) from source to target
                segment = nx.shortest_path(core_graph, source_node, target_node, weight='weight', method='dijkstra') # bellman-ford or dijkstra

                # Get length of path
                segment_length = len(segment)

                # Get length of segment with multi nodes removed
                two_degree_segment_length = len([node for node in segment if node not in multi_edge_nodes])

                # Check if no node between the source and target has more than two edges,
                # if then move to record the segment/path
                if segment_length - 2 == two_degree_segment_length:
                    # Check if two gene segment occur in every possible genome, if not then skip
                    if segment_length == 2:
                        gene_co_occurrences, _ = count_gene_co_occurrence(core_gene_dict, segment)
                        if num_gffs - core_graph[segment[0]][segment[1]]['weight'] < gene_co_occurrences:
                            continue
                        else:
                            if all([x != segment[::-1] for x in multi_edge_connect_adjust]): multi_edge_connect_adjust.append(segment)

                    # Construct name for path
                    source_target_name = sorted([source_node, target_node])
                    source_target_name = f'{source_target_name[0]}--{source_target_name[1]}'

                    # Check that path has not been recorded in the opposite direction, if not then record it
                    if source_target_name not in double_edge_segements:
                        double_edge_segements[source_target_name] = segment
                    else:
                        if double_edge_segements[source_target_name] != segment[::-1]:
                            exit_with_error(EXIT_SEGMENT_IDENTIFICATION_ERROR,
                                            f"Path from one node to another ({source_target_name}) was found, but did not match previously found path!")

    # Calculate the expected number of paths
    total_edges_from_multi_edge_nodes = sum([connections for _, connections in core_graph.degree if connections > 2])
    num_edges_between_multi_edge_nodes = sum([len(connect_dict[key]) for key in connect_dict])
    expected_segment_number = int((total_edges_from_multi_edge_nodes / 2) - (num_edges_between_multi_edge_nodes / 2)) + len(multi_edge_connect_adjust)

    # Check if less than the number of expected paths has been found,
    # if then try to identify missing paths
    if expected_segment_number != len(double_edge_segements):
        # Get number of expected edges for each node that has more than two edges
        expected_edge_num_dict = {node: connections for node, connections in core_graph.degree if connections > 2}

        # Get the number of edges directly between multi connected nodes
        identified_edge_num_dict = {node: len(connect_dict[node]) for node in connect_dict}

        # Get the number of paths connecting multi connected nodes via segments found in previous loop
        for connection in double_edge_segements:
            connection_nodes = connection.split('--')
            for node in connection_nodes:
                identified_edge_num_dict[node] += 1

        # Adjust for the number of segments identified between multi connected nodes
        for segment in multi_edge_connect_adjust:
            for node in segment:
                identified_edge_num_dict[node] -= 1

        # Compare the number of connections expected to the number identified, to find nodes that are miss connections
        nodes_missing_connections = []
        for node in expected_edge_num_dict:
            if identified_edge_num_dict[node] != expected_edge_num_dict[node]:
                nodes_missing_connections.append(node)

        # Go through nodes that are missing at least one path and try to identify missing paths
        for source_node in nodes_missing_connections:
            for target_node in nodes_missing_connections:
                # Check that the current target node is not a neighbouring node or the current node itself
                if target_node not in connect_dict[source_node] and target_node != source_node:
                    # Copy the graph to manipulate it
                    core_graph_copy = core_graph.copy()

                    # Construct a pair name
                    suspected_pair = sorted([source_node, target_node])
                    suspected_pair = f'{suspected_pair[0]}--{suspected_pair[1]}'

                    # Check that the pair has not been found in a previous run
                    if suspected_pair not in double_edge_segements:
                        # Counter to stop loop
                        counter = 0
                        # Identifier to see if path has been found, to stop loop
                        path_identified = False
                        while not path_identified:
                            counter += 1

                            # Get all shortest path between source and target.
                            all_shortest_paths = nx.all_shortest_paths(core_graph_copy, source_node, target_node)

                            # Go through each path to see if is satisfies the criteria
                            try:
                                for index, path in enumerate(all_shortest_paths):
                                    # Get length of path
                                    segment_length = len(path)

                                    # Get length of segment with multi nodes removed
                                    two_degree_segment_length = len([node for node in path if node not in multi_edge_nodes])

                                    # Check that the path does not contain nodes with >2 degrees outside of source and target,
                                    # if then add path,
                                    # else then find nodes that has >2 edges and remove an edge that leads to the node, to break the path for next run through loop
                                    if segment_length - 2 == two_degree_segment_length and two_degree_segment_length != 0:
                                        double_edge_segements[suspected_pair] = path
                                        path_identified = True
                                        continue
                                    else:
                                        # Check if path is length >2,
                                        # if then find >2 degree nodes and remove an edge to them,
                                        # else just remove edge found between nodes.
                                        if len(path) > 2:
                                            multi_node_in_path = [[path[index], path[index+1]] for index, node in enumerate(path) if node in multi_edge_nodes and node != source_node and node != target_node]
                                            for multi_node_pair in multi_node_in_path:
                                                # Try to remove edge found to multi node, if already removed move on.
                                                try:
                                                    core_graph_copy.remove_edge(*multi_node_pair)
                                                except nx.exception.NetworkXError:
                                                    continue
                                        else:
                                            core_graph_copy.remove_edge(*path)

                                    if counter == 1000:
                                        raise IndexError("Counter reached limit! in detecting a new path for pair.")
                            except nx.NetworkXNoPath:
                                # No simple paths could be found for the source and target thus the while loop is terminated.
                                path_identified = True

    return double_edge_segements


def determine_genome_segments(core_neighbour_pairs, combined_acc_gene_count, num_gffs, core_gene_dict, logger):
    """
    Function to be called from main that collects the functions for determining core segments in pan-genome

    :param core_neighbour_pairs: Dict of the number of times core pairs have been detected
    :param combined_acc_gene_count: Number of accessory and low-frequency genes detected between core gene pairs
    :param num_gffs: Number of inputted gff files # TODO - Should this be the minimum number determined by the input cut-off for core genes
    :param logger: Program logger

    :return double_edge_segements:
    :return no_acc_segments:
    """

    logger.debug(f"--------------Searching for segments in pan genome--------------")

    # Construct a graph from core gene neighbours
    core_graph = construct_core_graph(core_neighbour_pairs)

    # Find segments in the genome between core genes with multiple neighbors
    double_edge_segements = identify_segments(core_graph, num_gffs, core_gene_dict)

    if double_edge_segements is not None:
        logger.debug(f'A total of {len(double_edge_segements)} core genes were identified to have multiple neighbours.')
        logger.debug(f'Genes with multiple neighbours: {double_edge_segements}')

        logger.debug('Search for Segments with no accessory genes starts now')

        # Find segments of core genes with no accessory in between
        no_acc_segments = identify_no_accessory_segments(double_edge_segements, combined_acc_gene_count)

        logger.debug('Segments with no accessory genes is done')
    else:
        logger.debug(f'No segments can be identified in given pan-genome\n')
        no_acc_segments = None

    return double_edge_segements, no_acc_segments


if __name__ == '__main__':
    pass


import argparse
import sys

EXIT_COMMAND_LINE_ERROR = 2


def get_commandline_arguments(args):
    """
    Function that takes the input given to the commandline and passes it.
    will check for no input and '-help'
    :param args: List of input arguments given to the commandline
    :return: matched argument object for passing in main function.
    """
    # Set up parser
    parser = argparse.ArgumentParser(description='Welcome to Corekaburra!'
                                                 'An extension to pan-genome analyses that summarise genomic regions '
                                                 'between core genes and segments of neighbouring core genes using '
                                                 'gene synteny from a set of input genomes and a pan-genome folder.',
                                     add_help=False)

    required = parser.add_argument_group('Required arguments')
    run_mods = parser.add_argument_group('Analysis modifiers')
    output_control = parser.add_argument_group('Output control')
    rem_args = parser.add_argument_group('Other arguments')

    required.add_argument('-ig',
                          '--input_gffs',
                          help='Path to gff files used for pan-genome',
                          metavar='file.gff',
                          required=True,
                          dest='input_gffs',
                          nargs='+')

    required.add_argument('-ip',
                          '--input_pangenome',
                          help='Path to the folder produced by Panaroo or Roary',
                          metavar='path/to/pan_genome',
                          required=True,
                          dest='input_pan')

    run_mods.add_argument('-cg',
                          '--complete_genomes',
                          help='text file containing names of genomes that are to be handled as complete genomes',
                          required=False,
                          metavar='complete_genomes.txt',
                          default=None,
                          dest='comp_genomes')

    run_mods.add_argument('-a',
                          '--no_annotate_refound',
                          help='Flag to toggle off the creation of new gff files, with annotation of refound genes.\n'
                               'Only done if input pangenome is detected as coming from Panaroo',
                          required=False,
                          default=True,
                          action='store_false',
                          dest='annotate')

    run_mods.add_argument('-cc',
                          '--core_cutoff',
                          help='Percentage of isolates in which a core gene must be present [default: 1.0]',
                          metavar='1.0',
                          type=float,
                          default=1.0,
                          dest='core_cutoff')

    run_mods.add_argument('-lc',
                          '--low_cutoff',
                          help='Percentage of isolates where genes found in less than these are seen as low-frequency genes [default: 0.05]',
                          metavar='0.05',
                          type=float,
                          default=0.05,
                          dest='low_cutoff')

    output_control.add_argument('-o',
                                '--output',
                                help='Path to where output files will be placed [default: current folder]',
                                required=False,
                                type=str,
                                metavar='path/to/output',
                                default='.',
                                dest='output_path')

    output_control.add_argument('-p',
                                '--prefix',
                                help='Prefix for output files, if any is desired',
                                required=False,
                                default=None,
                                dest='output_prefix')

    output_control.add_argument('-d',
                                '--discard_corrected',
                                help='Discard gff files corrected with refound genes identified by Panaroo - Only compativle if pan-genome comes from Panaroo [Default: Corrected files are kept]',
                                required=False,
                                default=False,
                                action='store_true',
                                dest='discard_gffs')

    rem_args.add_argument('-c',
                          '--cpu',
                          help='Give max number of CPUs [default: 1]',
                          required=False,
                          metavar='int',
                          default=1,
                          type=int,
                          dest='cpu')

    logger_level = rem_args.add_mutually_exclusive_group()
    logger_level.add_argument('-l',
                              '--log',
                              help='Record program progress in for debugging purpose',
                              action='store_true',
                              default=False,
                              required=False)

    logger_level.add_argument('-q',
                              '--quiet',
                              help='Only print warnings',
                              action='store_true',
                              default=False,
                              required=False)

    rem_args.add_argument('-h',
                          '--help',
                          action='help',
                          help='Show help function')



    # Check if any thing is given as input otherwise warn and print help
    if len(args) < 1:
        parser.print_help()
        sys.exit(EXIT_COMMAND_LINE_ERROR)
    elif '-help' in args:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args(args)

    return args



from Bio import Seq
from Bio.SeqRecord import SeqRecord
import gffutils
from gffutils.gffwriter import GFFWriter
import os
import concurrent.futures
# from Bio.Blast import
from time import time

try:
    from Corekaburra.exit_with_error import exit_with_error
except ModuleNotFoundError:
    from exit_with_error import exit_with_error
EXIT_GFF_REANNOTATION_ERROR = 3


def read_gene_data(gene_data_file):
    """
    Function to read the gene_data.csv file outputted by Panaroo and
    :param gene_data_file: File path to the gene_data.csv file
    :return: A dict of genomes with their refound genes
    """

    # Construct dictionary to hold refound genes and sequences for these
    gene_data_dict = {}

    # Read the gene_data.csv file and record all refound genes
    with open(gene_data_file, 'r') as gene_data:

        for line in gene_data.readlines():
            # Split read line at commas
            line = line.split(',')

            # Check if refound gene
            if 'refound' in line[2]:
                # Try to add the refound gene to the gene_data dict as a second key, value being the DNA sequence,
                # if the first key (genome) is not found in gene_data dict,
                # then construct dict for the genome and add the gene
                try:
                    gene_data_dict[line[0]][line[2]] = [line[5], line[6], line[7].strip()]
                except KeyError:
                    gene_data_dict[line[0]] = {line[2]: [line[5], line[6], line[7].strip()]}

    return gene_data_dict


def prepair_for_reannotation(gene_data_path, output_folder, gffs, logger):
    """
    Function for creating an output folder for corrected genomes, check if any are present, and if then which.
    :param gene_data_path: Path to the gene_data.csv file from Panaroo
    :param output_folder: Folder designated as the output folder for Corekaburra
    :param gffs: List of file-paths to gff files.
    :param logger: Program logger

    :return gene_data_dict: Dict containing the information expected from the gene_data.csv file
    :return corrected_gff_out_dir: File path to the created or identified directory of corrected gff files
    :return gffs: List of gff files, some may be altered to be the corrected verison from prior runs/
    """

    logger.debug('Initialise structures for reannotating genes found by Panaroo')

    # Read Gene_data.csv file into dict with a dict of refound genes for each genome
    gene_data_dict = read_gene_data(gene_data_path)

    # Construct directory to hold corrected gff files:
    corrected_gff_out_dir = os.path.join(output_folder, 'Corrected_gff_files')
    # Try and construct folder,
    # if present check if content matches input to avoid process
    try:
        os.mkdir(corrected_gff_out_dir)
    except FileExistsError:
        corrected_folder_content = os.listdir(corrected_gff_out_dir)

        gff_names = [os.path.basename(gff) for gff in gffs]

        corrected_files = [file for file in corrected_folder_content if
                           f'{file.split("_corrected")[0]}.gff' in gff_names]

        if len(corrected_files) > 0:
            gffs = [file for file in gff_names if f'{file.replace(".gff", "")}_corrected.gff' not in corrected_files]
            gffs = gffs + corrected_files

    return gene_data_dict, corrected_gff_out_dir, gffs


def extract_genome_fasta(gff_name):
    """
    Function to read and extract information from a gff3 file
    :param gff_name: File path to a gff file

    :return genome_fasta_dict: Dict over contig names as keys, and values being the contig sequence
    :return largest_locus_tag: the largest locus_tag identified in gff file
    :return header_lines: the header lines proceeding annotations.
    """

    # Initialise the two return variables a dict for refound gene's annotaton and the largest locus_tag
    genome_fasta_dict = {}
    largest_locus_tag = ''
    header_lines = []

    # Open the gff file and indicate that the FASTA sequence has not beed reached
    with open(gff_name, 'r') as gff_file:
        found_fasta = False

        # Go through gff file and find and read the fasta seuqence at the end after the ##FASTA mark
        for line in gff_file.readlines():
            if found_fasta:
                # Check if line is fasta header,
                # If then construct new dict key if not append sequence to current key
                if '>' in line:
                    line = line.split(' ')[0]
                    try:
                        current_contig = line.strip()
                        current_contig = current_contig.split(">")[1]
                        genome_fasta_dict[current_contig] = ''
                    except KeyError:
                        raise KeyError("Some contig names contain redundant names when seperated by white space")
                else:
                    line = line.split('\n')[0]
                    genome_fasta_dict[current_contig] = genome_fasta_dict[current_contig] + line

            # Check if FASTA part of gff file has been found,
            # if then indicate to start recording sequences,
            # else compare locus_tag of the line
            elif '##FASTA' in line:
                found_fasta = True

            else:
                # Compare the locus_tag with the previously largest locus_tag,
                # save the largest of the two
                if '#' not in line:
                    line = line.split('\t')
                    line = line[8].split(';')
                    line = [element for element in line if 'locus_tag' in element]
                    # Examine non empty locus_tags
                    if len(line) > 0:
                        line = line[0]
                        line = line.split('locus_tag=')[1]
                        line = line.strip()
                        if line > largest_locus_tag:
                            largest_locus_tag = line
                # Save header lines
                else:
                    header_lines.append(line)

    return genome_fasta_dict, largest_locus_tag, header_lines


def add_gene_to_gff(tmp_gff, gene_oi, genome_oi, contig, strand, refound_gene_tag, annotation, largest_locus_tag):
    """
    Function to construct and append a line to a file.
    :param tmp_gff: An open file to append a line to
    :param gene_oi: Gene in question
    :param genome_oi: Genome in question
    :param contig: Contig of the genome
    :param strand: Strand of the gene in question
    :param annotation: Any annotation found in Panaroo
    :param refound_gene_tag: Tag given by Panaroo
    :param largest_locus_tag: The current largest locus_tag
    :return: The new largest locus_tag
    """
    gene_start = genome_oi.find(gene_oi) + 1
    gene_end = gene_start + len(gene_oi) - 1
    locus_tag_parts = largest_locus_tag.rsplit('_', maxsplit=1)
    tag_length = len(locus_tag_parts[1])

    locus_tag_parts[1] = int(locus_tag_parts[1]) + 1
    preceding_zeros = str(0) * (tag_length - len(str(locus_tag_parts[1])))
    # Add preceding zeros
    locus_tag_parts[1] = f'{preceding_zeros}{locus_tag_parts[1]}'
    new_locus_tag = f'{locus_tag_parts[0]}_{locus_tag_parts[1]}'
    description = annotation[1]
    name = annotation[0]

    # Construct gff field 9
    info_field = f'ID={new_locus_tag};locus_tag={new_locus_tag};old_locus_tag={refound_gene_tag}'

    if name != '':
        info_field += f';name={name}'
    if description != '':
        info_field += f';annotation={description}'

    # construct tab delimited string containing the features of the gene
    gff_line = f'{contig}\t' \
               f'Panaroo\t' \
               f'CDS\t' \
               f'{gene_start}\t' \
               f'{gene_end}\t' \
               f'.\t' \
               f'{strand}\t' \
               f'0\t' \
               f'{info_field}'

    tmp_gff.write(gff_line + '\n')

    return new_locus_tag


def write_contig(file, contig_name, sequence):
    """
    Write contig into file
    :param file: Open file for appending contig to
    :param contig_name: Name of the contig to be added
    :param sequence: Sequence of contig to be added
    :return: Nothing
    """
    # Write contig name
    file.write(f'>{contig_name}\n')

    # Write bulk of sequence
    for i in range(len(sequence) // 60):
        file.write(sequence[0+60*i:60+60*i] + '\n')

    # Write remainder of sequence
    remainder = len(sequence) % 60
    genome_length = len(sequence)
    file.write(sequence[len(sequence) - remainder:genome_length+1] + '\n')


def annotate_refound_genes(gff_name, gene_data_dict, tmp_folder_path, corrected_gff_out_dir, logger):
    """
    Function to add back in genes that are refound by Panaroo into gff files.
    :param gff_name: File path of gff to be corrected
    :param gene_data_dict: Dict of refound genes identified from gene_presence_absence.csv file
    :param tmp_folder_path: File path to the temporary folder
    :param corrected_gff_out_dir: File path to the folder where corrected genomes should be place
    :param logger: Program logger
    :return: Nothing.
    """
    """ Function to annotate the genes refound by Panaroo in a gff3 file"""
    # Read in a gff file
    # Get base name of gff file and construct path to database in temporary folder
    gff_file_name = os.path.basename(gff_name)
    data_base = os.path.join(tmp_folder_path, f'{gff_file_name}_db')

    # Create a database for the gff file
    gffutils.create_db(gff_name, data_base)
    # Attach database
    gff_db = gffutils.FeatureDB(data_base)

    # Write quick tmp database for appending new genes.
    tmp_gff = os.path.join(tmp_folder_path, f'{gff_file_name.split(".gff")[0]}_tmp.gff')
    with open(tmp_gff, 'w') as gff_file:
        for feature in gff_db.all_features():
            gff_file.writelines(str(feature) + '\n')

    # Pass the gff file manually to extract the genome fasta sequence(s) and the largest locus_tag
    fasta_genome, largest_locus_tag, header_lines = extract_genome_fasta(gff_name)

    # Find all refound genes for given genome in gene data file
    genome_name = gff_file_name.split('.')[0]

    # Search for the refound genes and record their coordinate, strand and add them to the gff file
    with open(tmp_gff, 'a') as tmp_gff_file:
        for refound_gene in gene_data_dict[genome_name]: # .keys()
            gene_oi = gene_data_dict[genome_name][refound_gene][0]

            strand = None
            contig_counter = 0
            contigs = list(fasta_genome) # .keys()
            while strand is None and contig_counter < len(contigs) :
                contig = contigs[contig_counter]
                genome_oi = fasta_genome[contig]

                if gene_oi in genome_oi:
                    strand = '+'

                else:
                    # get reverse complement of the gene
                    gene_oi = Seq.reverse_complement(gene_oi)
                    if gene_oi in genome_oi:
                        strand = '-'

                contig_counter += 1

            if strand is not None:
                # Add the gene to the gff file.
                largest_locus_tag = add_gene_to_gff(tmp_gff_file, gene_oi, genome_oi, contig, strand,
                                                    refound_gene, gene_data_dict[genome_name][refound_gene][1:], largest_locus_tag)
            else:
                exit_with_error(f"When correcting gff {gff_name}, the gene: {refound_gene} "
                                 f"did not have any hit in the genome!", EXIT_GFF_REANNOTATION_ERROR, logger)

    # Construct a database from the temporary gff that contain the added annotations
    path_tmp_gff_db = os.path.join(tmp_folder_path, f'{gff_file_name}_tmp_db')
    # make database
    gffutils.create_db(tmp_gff, path_tmp_gff_db)
    # Attach database
    tmp_gff_db = gffutils.FeatureDB(path_tmp_gff_db)

    # Print the final GFF3 file
    corrected_gff_file = os.path.join(corrected_gff_out_dir, f'{gff_file_name.split(".gff")[0]}_corrected.gff')
    with open(corrected_gff_file, 'w') as gff_file:
        # Write initial lines
        for line in header_lines:
            gff_file.write(line)

        # ADD the gff lines
        for feature in tmp_gff_db.all_features(order_by=('seqid', 'start')):
            gff_file.writelines(str(feature) + '\n')

        # Write line to separate genome fasta
        gff_file.write("##FASTA\n")

        # Write the genome fasta
        for contig_name in fasta_genome.keys():
            write_contig(gff_file, contig_name, fasta_genome[contig_name])

    # remove database for gff file and the temporary gff file in temporary folder
    os.remove(data_base)
    os.remove(tmp_gff)
    os.remove(path_tmp_gff_db)

    return corrected_gff_file


if __name__ == '__main__':
    pass
    # _, _, attribute_dict = read_gene_presence_absence('/Users/mjespersen/OneDrive - The University of Melbourne/Phd/Parts/Recombination_hotspots/Code/Between_core_variation/data_for_unit_tests/test_pan_genome/50_refseq_pan_split_paralogs/gene_presence_absence_roary.csv',
    #                                                   1, 0.05)
    #
    # correct_gffs(['/Users/mjespersen/OneDrive - The University of Melbourne/Phd/Parts/Recombination_hotspots/Code/Between_core_variation/data_for_unit_tests/test_pan_genome/50_refseq_genomes/GCA_008694005.gff'], '/Users/mjespersen/OneDrive - The University of Melbourne/Phd/Parts/Recombination_hotspots/Code/Between_core_variation/data_for_unit_tests/test_pan_genome/50_refseq_pan_split_paralogs/gene_data.csv',
    #                  "/Users/mjespersen/OneDrive - The University of Melbourne/Phd/Parts/Recombination_hotspots/Code/Between_core_variation/data_for_unit_tests", attribute_dict)
    # # genome_dict = extract_genome_fasta('/Users/mjespersen/OneDrive - The University of Melbourne/Phd/Parts/Recombination_hotspots/Code/Between_core_variation/data_for_unit_tests/test_pan_genome/50_refseq_genomes/GCA_000006785.gff')



'''
Module      : Main
Description : The main entry point for the Corekaburra.
Copyright   : (c) Magnus Ganer Jespersen, 11 Oct 2021 
License     : MIT 
Maintainer  : magnus.ganer.j@gmail.com 
Portability : POSIX

Corekaburra looks at the gene synteny across genomes used to build a pan-genome. Using syntenic information Corekaburra
identifies regions between core gene clusters. Regions are described in terms of their content of accessory gene clusters
and distance between core genes. Information from neighboring core genes is further used to identify stretches of core
gene clusters throughout the pan-genome that appear in all genomes given as input. Corekaburra is compatible with outputs
from standard pan-genome pipelines: Roary and Panaroo.
'''

import os
import logging
import time
import concurrent.futures

try:
    from Corekaburra.commandline_interface import get_commandline_arguments
except ModuleNotFoundError:
    from commandline_interface import get_commandline_arguments

try:
    from Corekaburra.read_complete_genome_file import parse_complete_genome_file
except ModuleNotFoundError:
    from read_complete_genome_file import parse_complete_genome_file

try:
    from Corekaburra.check_inputs import define_pangenome_program, check_gene_data, check_gff_in_pan, check_cutoffs
except ModuleNotFoundError:
    from check_inputs import define_pangenome_program, check_gene_data, check_gff_in_pan, check_cutoffs

try:
    from Corekaburra.parse_gene_presence_absence import read_gene_presence_absence
except ModuleNotFoundError:
    from parse_gene_presence_absence import read_gene_presence_absence

try:
    from Corekaburra.correct_gffs import prepair_for_reannotation
except ModuleNotFoundError:
    from correct_gffs import prepair_for_reannotation

try:
    from Corekaburra.gff_parser import segment_genome_content
except ModuleNotFoundError:
    from gff_parser import segment_genome_content

try:
    from Corekaburra.merge_dicts import merge_dicts_lists, merge_dicts_counts
except ModuleNotFoundError:
    from merge_dicts import merge_dicts_lists, merge_dicts_counts

try:
    from Corekaburra.consesus_core_genome import determine_genome_segments
except ModuleNotFoundError:
    from consesus_core_genome import determine_genome_segments

try:
    from Corekaburra.summary_table import calculate_n_create_summaries
except ModuleNotFoundError:
    from summary_table import calculate_n_create_summaries

try:
    from Corekaburra.output_writer_functions import master_info_writer, summary_info_writer, segment_writer, no_acc_segment_writer, non_core_contig_writer
except ModuleNotFoundError:
    from output_writer_functions import master_info_writer, summary_info_writer, segment_writer, no_acc_segment_writer, non_core_contig_writer

import sys
import pkg_resources

EXIT_INPUT_FILE_ERROR = 1
EXIT_COMMAND_LINE_ERROR = 2
EXIT_GFF_REANNOTATION_ERROR = 3
EXIT_SEGMENT_IDENTIFICATION_ERROR = 4
DEFAULT_VERBOSE = False
PROGRAM_NAME = "Corekaburra"


try:
    PROGRAM_VERSION = pkg_resources.require(PROGRAM_NAME)[0].version
except pkg_resources.DistributionNotFound:
    PROGRAM_VERSION = "undefined_version"


def init_logging(debug_log, quiet, out_path):
    """
    initialise the logging file, and write log statement
    indicating the program has started, and also write out the
    command line from sys.argv
    :param debug_log: Bool indicating if log is a debug log
    :param quiet: Bool indicating if logging should be kept minimal
    :param out_path: Output path for the program, and where log will be placed
    :return: Logger object
    """
    if debug_log:
        level = logging.DEBUG
    elif quiet:
        level = logging.WARNING
    else:
        level = logging.INFO

    # Construct logger logging to file
    file_logger = logging.getLogger(__name__)
    file_logger.setLevel(level)

    formatter = logging.Formatter('[%(asctime)s] - %(levelname)s - %(module)s: %(message)s',
                                  datefmt="%Y-%m-%d %H:%M:%S%z")

    file_handler = logging.FileHandler(os.path.join(out_path, 'Corekaburra.log'))
    file_handler.setLevel(level)
    file_handler.setFormatter(formatter)
    file_logger.addHandler(file_handler)

    # Log command-line argument and debug line for Corekaburra start
    file_logger.info(f"command line: {' '.join(sys.argv)}")

    return file_logger


def stream_logging(file_logger):
    """
    Function adding in stream logging following initial logging
    :param file_logger: Logger object
    :return: Logger object with added stream logging
    """
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)

    file_logger.addHandler(stream_handler)

    file_logger.info('\n----------------------Processing started----------------------\n')

    return file_logger


def main():
    """
    This is the main function for running Corekaburra.
    Requires a pan-genome folder from an allowed program, along with GFF3 files used for producing the pan-genome
    """
    total_time_start = time.time()

    inital_check_time_start = time.time()

    # get arguments from the commandline
    args = get_commandline_arguments(sys.argv[1:])

    # Construct output folder
    try:
        os.mkdir(args.output_path)
    except FileExistsError:
        pass

    # Run initialisation of logger:
    logger = init_logging(args.log, args.quiet, args.output_path) # TODO - if not dependency check is done then it should be possible to add the stream logger following the logging of the command line in the initial logging function.
    logger = stream_logging(logger)

    # Check that low-frequency cutoff and core cutoff are as expected
    check_cutoffs(args.low_cutoff, args.core_cutoff, logger)

    # TODO - Make Corekaburra take gzipped inputs
    # TODO - Add so that a single gff file can only be given as input once and not multiple times?

    # Check the presence of provided complete genomes among input GFFs
    if args.comp_genomes is not None:
        comp_genomes = parse_complete_genome_file(args.comp_genomes, args.input_gffs, logger)
    else:
        comp_genomes = None

    # Check if Panaroo or Roary input folder is given
    source_program, input_pres_abs_file_path = define_pangenome_program(args.input_pan, logger)

    # Check if gene_data file is present if Panaroo input is given an gffs should be annotated
    if args.annotate and source_program == 'Panaroo':
        gene_data_path = check_gene_data(args.input_pan, logger)
    else:
        gene_data_path = None

    check_gff_in_pan(args.input_gffs, input_pres_abs_file_path, logger)

    # Construct temporary folder:
    # TODO - check that the temporary folder does not exist and that the user does not have a folder with same name already. (Maybe use a time stamp for the start to make it unique.)
    tmp_folder_path = os.path.join(args.output_path, 'Corekaburra_tmp')
    try:
        os.mkdir(tmp_folder_path)
    except FileExistsError:
        for file in os.listdir(tmp_folder_path):
            os.remove(file)

    logger.info('Initial checks successful\n')
    inital_check_time_end = time.time()

    ## Read in gene presence absence file
    time_start_read_files = time.time()
    # TODO - ATM the column with presence of gene in genomes is used to define what is core and not. Is it better to use the number of input gffs instead?
    #   - There are upsides to the current. You can use the same genome to find segments for two different populations with in the dataset using the same reference of core-genes
    #   - Making it depend on the input is not viable for comparing runs, even within the same pan-genome, when using different sets of gff files.
    # TODO - Some day it would be awesome to be able to provide a clustering/population structure which could divide genes into the 13 definitions outlined by Horesh et al. [DOI: 10.1099/mgen.0.000670]
    core_dict, low_freq_dict, acc_gene_dict = read_gene_presence_absence(input_pres_abs_file_path, args.core_cutoff,
                                                                         args.low_cutoff, source_program,
                                                                         args.input_gffs, tmp_folder_path, logger)

    # Prepair folder for reannotated genes and examine if any are already present
    if source_program == "Panaroo" and args.annotate:
        gene_data_dict, corrected_dir, args.input_gffs = prepair_for_reannotation(gene_data_path, args.output_path,
                                                                                   args.input_gffs, logger)
    else:
        gene_data_dict = None
        corrected_dir = None

    time_end_read_files = time.time()
    time_start_passing_gffs = time.time()

    # Loop over all Gffs and extract info from each of them.
    time_start = time.time()
    # Initialise dictionaries to contain results from all gff files
    core_neighbour_pairs = {}
    core_neighbour_distance = {}
    core_neighbour_accessory_count = {}
    core_neighbour_low_freq = {}
    master_info_total = {}
    non_core_contig_info = {}

    progress_counter = 0
    if len(args.input_gffs) > 10:
        progress_update = len(args.input_gffs) / 10
    else:
        progress_update = 1

    with concurrent.futures.ProcessPoolExecutor(max_workers=args.cpu) as executor:
        logger.info(f"------Start core region identification of given gff files-----\n")
        logger.info(f'{len(args.input_gffs)} GFF files to process')

        results = [executor.submit(segment_genome_content, gff, core_dict, low_freq_dict, acc_gene_dict, comp_genomes,
                                   source_program, args.annotate, gene_data_dict, corrected_dir, tmp_folder_path, args.discard_gffs, logger)
                   for gff in args.input_gffs]

        for output in concurrent.futures.as_completed(results):
            progress_counter += 1
            if progress_counter % progress_update == 0 or progress_counter == 1:
                logger.info(f"GFF file #{progress_counter} has been processed")

            # Split the outputs
            core_pairs, distance, acc_count, \
            low_freq, master_info_return, \
            core_less_contigs_return = output.result()

            # Merge results into single/master dictionaries
            core_neighbour_pairs = merge_dicts_counts(core_neighbour_pairs, core_pairs)
            core_neighbour_distance = merge_dicts_lists(core_neighbour_distance, distance)
            core_neighbour_accessory_count = merge_dicts_lists(core_neighbour_accessory_count, acc_count)
            core_neighbour_low_freq = merge_dicts_lists(core_neighbour_low_freq, low_freq)
            master_info_total.update(master_info_return)
            non_core_contig_info.update(core_less_contigs_return)

    time_end_passing_gffs = time.time()
    time_start_segments_search = time.time()

    time_start = time.time()
    # Count number of unique accessory genes inserted into a core-core region across the genomes
    acc_region_count = {key: len(set(core_neighbour_low_freq[key])) for key in core_neighbour_low_freq}
    # Count number of unique low frequency genes inserted into a core-core region across the genomes
    low_frew_region_count = {key: len(set(core_neighbour_accessory_count[key])) for key in
                             core_neighbour_accessory_count}

    # Combine the accessory and low frequency counts:
    combined_acc_gene_count = {key: low_frew_region_count[key] + acc_region_count[key] for key in low_frew_region_count}

    double_edge_segements, no_acc_segments = determine_genome_segments(core_neighbour_pairs, combined_acc_gene_count,
                                                                       len(args.input_gffs), core_dict, logger)

    time_end_segments_search = time.time()

    # Produce dict containing summarised information from master info.
    logger.debug("Commence on calculating summary output")
    master_summary_info = calculate_n_create_summaries(master_info_total, core_dict)

    ### WRITE OUTPUTS ###
    logger.debug("-----------------------Printing outputs-----------------------")
    # Write master information to output file
    time_start = time.time()
    logger.debug("Master outputs")
    master_info_writer(master_info_total, args.output_path, args.output_prefix)

    logger.debug("Summary output")
    summary_info_writer(master_summary_info, args.output_path, args.output_prefix)

    if double_edge_segements is not None:
        logger.debug("Segment output")
        segment_writer(double_edge_segements, args.output_path, args.output_prefix)

        logger.debug("No Accessory segment output")
        no_acc_segment_writer(no_acc_segments, args.output_path, args.output_prefix)
    # TODO - Possibly output core gene graph. with segment annotations in colour - possibly info on edges using weight for conenctions and other atributes for acc content.?

    # TODO - Print summary number of genes and names for non-core contigs
    # TODO - Should we print a low-freq, placement?
    if len(non_core_contig_info)> 0:
        logger.debug("Non-core contig output")
        non_core_contig_writer(non_core_contig_info, args.output_path, args.output_prefix)

    # time_calculator(time_start, time.time(), "writing output files")

    # Finish up running
    total_time = round(time.time() - total_time_start, 1)
    initial_time = round(inital_check_time_end - inital_check_time_start, 1)
    read_fies_time = round(time_end_read_files - time_start_read_files, 1)
    passing_gffs_time = round(time_end_passing_gffs - time_start_passing_gffs, 1)
    segment_search_time = round(time_end_segments_search - time_start_segments_search)

    logger.debug("-----------------------Time used in run-----------------------")
    logger.debug(f"Total time used: {total_time}s")
    logger.debug(f"Initial check time: {initial_time}s")
    logger.debug(f"Reading pan-genome files time: {read_fies_time}s")
    logger.debug(f"Passing over Gff files time: {passing_gffs_time}s")
    logger.debug(f"Searching for segments time: {segment_search_time}s")


    # Remove temporary database holding gff databases
    if os.path.isdir(tmp_folder_path):
        os.rmdir(tmp_folder_path)
    if args.discard_gffs:
        os.rmdir(os.path.join(args.output_path, 'Corrected_gff_files'))


if __name__ == '__main__':
    main()



#!/usr/bin/env python

from distutils.core import setup

LONG_DESCRIPTION = \
'''Corekaburra looks at the gene synteny across genomes used to build a pan-genome. Using syntenic information Corekaburra 
identifies regions between core gene clusters. Regions are described in terms of their content of accessory gene clusters 
and distance between core genes. Information from neighboring core genes is further used to identify stretches of core  
gene clusters throughout the pan-genome that appear in all genomes given as input. Corekaburra is compatible with outputs 
from standard pan-genome pipelines: Roary and Panaroo.'''


setup(
    name='Corekaburra',
    version='0.0.1',
    author='Magnus Ganer Jespersen',
    author_email='magnus.ganer.j@gmail.com',
    packages=['Corekaburra'],
    package_dir={'Corekaburra': 'Corekaburra'},
    entry_points={
        'console_scripts': ['Corekaburra = Corekaburra.__main__:main']
    },
    url='https://github.com/milnus/Corekaburra',
    license='LICENSE',
    description=('A prototypical bioinformatics command line tool'),
    long_description=(LONG_DESCRIPTION),
    install_requires=["biopython", "networkx", "gffutils", "numpy"],
)



import os

try:
    from Corekaburra.correct_gffs import annotate_refound_genes
except ModuleNotFoundError:
    from correct_gffs import annotate_refound_genes


def parse_gff(input_file):
    """
    Read a gff file and return it as a generator object that return all line containing CDS
    :param input_file: File-path to a given gff file to be processed
    :return: Generator object returning CDS from a gff file
    """
    with open(input_file, 'r') as gff_file:
        for line in gff_file:
            if "##FASTA" in line:
                break
            if "#" not in line and 'CDS' in line:
                # Strip line for newline and split columns into list
                line = line.strip()
                line = line.split("\t")

                # See if refound gene or Prokka annotated and isolate ID in gene_presence_absence.csv accordingly
                if "old_locus_tag=" in line[8]:
                    gene_id = line[8][line[8].find('old_locus_tag'):]
                    if ';' in gene_id:
                        gene_id = gene_id[:gene_id.find(';')]
                else:
                    gene_id = line[8][line[8].find('ID'):line[8].find(';')]

                # Remove equal sign from id and add as identifier for the returned gff line
                gene_id = gene_id[gene_id.find('=') + 1:]
                line[8] = gene_id
                yield line


def get_contig_lengths(input_file):
    """
    Function that takes an input gff file path and records the length of each contig in the file
    :param input_file: File path for a given gff file
    :return: directory with key being the contig name (before first white spae)
    and value the size of the contig in base pairs
    """
    fasta_reached = False
    contig_size = 0
    contig_size_dir = {}

    # Open the given gff file, find the fasta section of the file and count the length of each contig.
    with open(input_file, 'r', ) as gff_file:
        for line in gff_file:
            if fasta_reached and '>' not in line:
                contig_size += len(line.rstrip())
            if fasta_reached and '>' in line:
                if contig_size > 0:
                    # Record the previous contig
                    if contig_name not in contig_size_dir:
                        contig_size_dir[contig_name] = contig_size
                    else:
                        raise ValueError(f"contig name: {contig_name}, in file {input_file} is duplicated! Please fix this")

                    # Set the contig name to the next contig
                    contig_name = line.strip().split(' ')[0].replace('>', '')
                    contig_size = 0
                else:
                    contig_name = line.strip().split(' ')[0].replace('>', '')

            if "##FASTA" in line:
                fasta_reached = True

        # Record last contig
        contig_size_dir[contig_name] = contig_size

    return contig_size_dir


def record_core_core_region(core_genes, gff_name, gff_line, contig_end, previous_core_gene_id,
                            previous_core_gene_end_coor, acc_genes_in_region, low_freq_genes_in_region,
                            core_gene_pair_distance, accessory_gene_content,
                            low_freq_gene_content, core_gene_pairs, master_info):
    """
    Function to record information about a core gene pair or a core gene and a sequence break,
    along with accessory information between the two features
    :param core_genes: Dict of core genes passed to genomes and the pan-genome clusters.
    :param gff_name: Name of the gff file currently being examined
    :param gff_line: List for the gene currently being recorded. (Can be None, when previous gene is next to sequence break)
    :param contig_end: Length of the contig in question, used to calculate the distance from last gene to end of contig
    :param previous_core_gene_id: ID of the previous gene recorded in pair, can be Sequence_break
    :param previous_core_gene_end_coor: End coordinate of the previously recorded genome (Can be None when new contig is initiated)
    :param acc_genes_in_region: List of recorded accessory-frequency genes in the region being recorded
    :param low_freq_genes_in_region: List of recorded low-frequency genes in the region being recorded
    :param core_gene_pair_distance: Dict of distances between core pairs recorded. key being the pair and value the distance in base-pairs
    :param accessory_gene_content: Dict of accessory frequency genes and their mapping to genomes
    :param low_freq_gene_content: Dict of low frequency genes and their mapping to genomes
    :param core_gene_pairs: List of core pairs recorded
    :param master_info: Large dict holding key for each pair recorded with a list of info around the pair as value
    :return: A tuple of dictionaries and lists, most of which are also given as input - See descriptions above
    """

    # Check that a line from gff is provided and previous gene is not a sequence break
    if gff_line is not None and previous_core_gene_id != "Sequence_break":
        # Check if core gene is fragmented
        if core_genes[gff_name][previous_core_gene_id] == core_genes[gff_name][gff_line[8]]:
            previous_core_gene_id = gff_line[8]
            previous_core_gene_end_coor = int(gff_line[4])
            acc_genes_in_region = []
            low_freq_genes_in_region = []
            return (previous_core_gene_id, previous_core_gene_end_coor, acc_genes_in_region, low_freq_genes_in_region,
                    core_gene_pair_distance, accessory_gene_content, low_freq_gene_content, core_gene_pairs, master_info)

    # Set core cluster names
    # If no line from gff is given there is a sequence break,
    # if it is given then set current cluster and try to find previous if not found it is a sequence break
    if gff_line is not None:
        current_core_gene_cluster = core_genes[gff_name][gff_line[8]]
        try:
            previous_core_gene_cluster = core_genes[gff_name][previous_core_gene_id]
        # Catch is previous gene was a sequence break.
        except KeyError:
            previous_core_gene_cluster = previous_core_gene_id

        core_gene_neighbours = sorted([previous_core_gene_cluster, current_core_gene_cluster])

    else:
        current_core_gene_cluster = "Sequence_break"
        previous_core_gene_cluster = core_genes[gff_name][previous_core_gene_id]
        core_gene_neighbours = [previous_core_gene_cluster, current_core_gene_cluster]

    # Merge core neighbours and add to pairs
    core_gene_neighbours_str = f'{core_gene_neighbours[0]}--{core_gene_neighbours[1]}'
    core_gene_pairs.append(core_gene_neighbours_str)

    # Set core neighbour distance
    # Check if measuring between two genes on same contig and not measuring from a sequence break
    if gff_line is not None and previous_core_gene_cluster != "Sequence_break":
        core_core_distance = int(gff_line[3]) - previous_core_gene_end_coor - 1
    else:
        # Check if measuring between sequence break and first core gene on contig, if then set start coordinate to zero
        if previous_core_gene_id == "Sequence_break":
            core_core_distance = int(gff_line[3]) - 1
            # If gene starts at contig end, make distance zero
            if core_core_distance < 0:
                core_core_distance = 0
        elif current_core_gene_cluster == 'Sequence_break':
            core_core_distance = abs(contig_end - previous_core_gene_end_coor)
        else:
            NotImplementedError(
                'An error occured when measuring the distance between core gene and contig end. Something went wrong!')

    # Add core neighbour distance
    core_gene_pair_distance[core_gene_neighbours_str] = core_core_distance

    # Add counts and annotation for accessory and low frequency genes
    acc_genes_in_region = sorted(list(set(acc_genes_in_region)))
    low_freq_genes_in_region = sorted(list(set(low_freq_genes_in_region)))

    accessory_gene_content[core_gene_neighbours_str] = acc_genes_in_region.copy()
    low_freq_gene_content[core_gene_neighbours_str] = low_freq_genes_in_region.copy()

    # Add info to master dict
    master_info[f'{core_gene_neighbours_str}--{gff_name}'] = [gff_name,
                                                              core_gene_neighbours[0],
                                                              core_gene_neighbours[1],
                                                              core_core_distance,
                                                              len(acc_genes_in_region) + len(low_freq_genes_in_region),
                                                              acc_genes_in_region.copy(),
                                                              low_freq_genes_in_region.copy()]

    # Update previous core gene id and end of core gene
    if gff_line is not None:
        # if previous_core_gene_id is not "Sequence_break":
        previous_core_gene_id = gff_line[8]
        previous_core_gene_end_coor = int(gff_line[4])
    else:
        previous_core_gene_id = "Sequence_break"
        # previous_core_gene_end_coor = 0

    # Reset values for accessory genes next core-core region
    acc_genes_in_region = []
    low_freq_genes_in_region = []

    return (previous_core_gene_id, previous_core_gene_end_coor, acc_genes_in_region, low_freq_genes_in_region,
            core_gene_pair_distance, accessory_gene_content, low_freq_gene_content, core_gene_pairs, master_info)


def connect_first_n_last_gene_on_contig(core_genes, gff_name, previous_core_gene_id, previous_core_gene_end_coor,
                                        first_core_gene_gff_line, acc_genes_in_region, first_core_accessory_content,
                                        low_freq_genes_in_region, first_core_low_freq_genes, contig_size,
                                        core_gene_pairs, core_gene_pair_distance,
                                        accessory_gene_content, low_freq_gene_content, master_info):
    """
    Function to record the connection between the first and the last genome on a closed contig

    :param core_genes: A dict of dicts mapping genomes to gene IDs to pan-genome clusters of core genes.
    :param gff_name: Name of the gff file currently being examined
    :param previous_core_gene_id: ID of the previous gene recorded in pair, can be Sequence_break.
    :param previous_core_gene_end_coor: End coordinate of the previously recorded genome.
    :param first_core_gene_gff_line: List for the gene first encountered on the contig.
    :param acc_genes_in_region: List of recorded accessory-frequency genes in the region after last core genes
    :param first_core_accessory_content: List of recorded accessory-frequency genes in the region before first core genes
    :param low_freq_genes_in_region: List of recorded low-frequency genes in the region after last core genes
    :param first_core_low_freq_genes: List of recorded low-frequency genes in the region before first core genes
    :param contig_size: Size of the contig currently being looked at.
    :param core_gene_pairs: List of core pairs recorded
    :param core_gene_pair_distance: Dict of distances between core pairs recorded. key being the pair and value the distance in base-pairs
    :param accessory_gene_content: Dict of accessory frequency genes and their mapping to genomes
    :param low_freq_gene_content: Dict of low frequency genes and their mapping to genomes
    :param master_info: Large dict holding key for each pair recorded with a list of info around the pair as value

    :return previous_core_gene_id: String of the last core genes ID/locus_tag
    :return previous_core_gene_end_coor: Int of the end coordinate for the latest core gene
    :return acc_genes_in_region: Empty list to store accessory genes
    :return low_freq_genes_in_region: Empty list to store low-frequency genes
    :return core_gene_pairs: List of core gene pairs recorded
    :return core_gene_pair_distance: A dict of distances between a specific pair of core genes.
    :return accessory_gene_content: A dict of the accessory genes found between a pair of core genes.
    :return low_freq_gene_content: A dict of the low-frequency genes found between a pair of core genes.
    :return master_info: A dict of multiple pieces of info for each core gene pair.
    """
    last_core_gene_cluster = core_genes[gff_name][previous_core_gene_id]
    first_core_gene_cluster = core_genes[gff_name][first_core_gene_gff_line[8]]


    # Add core neighbours
    core_gene_neighbours = sorted([last_core_gene_cluster, first_core_gene_cluster])
    core_gene_neighbours_str = f'{core_gene_neighbours[0]}--{core_gene_neighbours[1]}'
    core_gene_pairs.append(core_gene_neighbours_str)

    # Add core neighbour distance
    core_core_distance = contig_size - previous_core_gene_end_coor + int(first_core_gene_gff_line[3]) - 1

    core_gene_pair_distance[core_gene_neighbours_str] = core_core_distance

    # Add accessory information from between last and first core gene
    last_first_accessory_content = acc_genes_in_region.copy() + first_core_accessory_content.copy()
    last_first_low_freq_count = low_freq_genes_in_region.copy() + first_core_low_freq_genes.copy()

    last_first_accessory_content = sorted(list(set(last_first_accessory_content)))
    last_first_low_freq_count = sorted(list(set(last_first_low_freq_count)))

    accessory_gene_content[core_gene_neighbours_str] = last_first_accessory_content.copy()
    low_freq_gene_content[core_gene_neighbours_str] = last_first_low_freq_count.copy()

    # Add to master info dict
    master_info[f'{core_gene_neighbours_str}--{gff_name}'] = [gff_name,
                                                              core_gene_neighbours[0],
                                                              core_gene_neighbours[1],
                                                              core_core_distance,
                                                              len(last_first_accessory_content) +
                                                              len(last_first_low_freq_count),
                                                              last_first_accessory_content,
                                                              last_first_low_freq_count]

    previous_core_gene_id = ""
    previous_core_gene_end_coor = int(first_core_gene_gff_line[4])
    acc_genes_in_region = []
    low_freq_genes_in_region = []

    return (previous_core_gene_id, previous_core_gene_end_coor, acc_genes_in_region, low_freq_genes_in_region,
            core_gene_pairs, core_gene_pair_distance, accessory_gene_content, low_freq_gene_content, master_info)


def record_coreless_contig(coreless_contigs, acc_genes_in_region, low_freq_genes_in_region, gff_name, contig_name):
    acc_genes_in_region = sorted(list(set(acc_genes_in_region)))
    low_freq_genes_in_region = sorted(list(set(low_freq_genes_in_region)))
    if len(acc_genes_in_region) + len(low_freq_genes_in_region) > 0:
        coreless_contigs[f'{gff_name}--{contig_name}'] = [acc_genes_in_region, low_freq_genes_in_region]

    return coreless_contigs


def segment_gff_content(gff_generator, core_genes, low_freq_genes, gff_path, acc_genes, complete_genomes):
    """
    Function that takes a gff generator, core, accessory, and low frequency genes and identify each core-core gene region
    counts the number of accessory genes in the region, records the number of low frequency genes in the region,
    and records the distance from one core gene to the next.
    :param gff_generator: A generator object providing each CDS line from a gff file to be segmented.
    :param core_genes: A dict of dicts mapping genomes to gene IDs to pan-genome clusters of core genes.
    :param low_freq_genes: Same structure as core_genes, but for low-frequency genes.
    :param gff_path: List of file paths to gff files.
    :param acc_genes: Same structure as core_genes, but for accessory genes.
    :param complete_genomes: List of gff names that are to be handled as complete.

    :return core_gene_pairs: A list of core genes (or sequence breaks) that are found to be neighbouring each other in a given gff file.
    :return core_gene_pair_distance: A dict of distances between a specific pair of core genes.
    :return accessory_gene_content: A dict of the accessory genes found between a pair of core genes.
    :return low_freq_gene_content: A dict of the low-frequency genes found between a pair of core genes.
    :return master_info: A dict of multiple pieces of info for each core gene pair (Gff file, core gene 1, core gene 2, distnace between them, genes between them, list of accesspry genes, list of low-frequency genes)
    :return coreless_contigs: Dict of contigs found to not encode any core genes on them. The accessory and low-frequency genes are recorded.
    """

    # Initialize data structures to be returned
    core_gene_pairs = []
    core_gene_pair_distance = {}
    low_freq_gene_content = {}
    accessory_gene_content = {}
    master_info = {}
    coreless_contigs = {}

    # Examine if a complete genome has been given
    if complete_genomes is None:
        complete_genome = False
    else:
        if os.path.basename(gff_path).replace('.gff', '').replace('_corrected', '') in complete_genomes:
            complete_genome = True
        else:
            complete_genome = False

    # Get size of contigs for given gff
    contig_sizes = get_contig_lengths(gff_path)

    # Split input path of gff to get genome name
    gff_name = gff_path.split('/')[-1]
    gff_name = gff_name.replace('.gz', '').rsplit('.', 1)[0]
    if 'corrected' in gff_name:
        gff_name = gff_name.split('_corrected')[0]

    # Set that first core gene has not been found
    first_core_gene = True
    # Set Initialise variable for the ID of previous encountered core gene
    previous_core_gene_id = ""
    # Set variable to get first contig
    first_contig = True
    # Initialise the accessory and low-frequency gene lists
    low_freq_genes_in_region = []
    acc_genes_in_region = []

    # Go through each line of GFF file
    for line in gff_generator:
        # Set first contig to fund in file
        if first_contig:
            previous_contig = line[0]
            first_contig = False

        # Check if contig has changed - if then finish contig, if not examine next gene on contig
        if line[0] == previous_contig:
            # Check if gene is core, if then test if it is the first, else assume to be accessory gene
            if line[8] in core_genes[gff_name]:

                # Check if core gene is the first observed in file - if then set information else record information
                if first_core_gene:
                    # Set information on first core gene to be used when finishing search
                    first_core_gene_gff_line = line
                    first_core_accessory_content = acc_genes_in_region.copy()
                    first_core_low_freq_genes = low_freq_genes_in_region.copy()

                    # Set information to be used with next core gene neighbour
                    previous_core_gene_end_coor = int(line[4])
                    previous_core_gene_id = line[8]

                    # Set that first core gene has been observed
                    first_core_gene = False

                    # Reset accessory and low frequency gene counters
                    low_freq_genes_in_region = []
                    acc_genes_in_region = []

                else:
                    # Record core gene pair information
                    (previous_core_gene_id,
                     previous_core_gene_end_coor,
                     acc_genes_in_region,
                     low_freq_genes_in_region,
                     core_gene_pair_distance,
                     accessory_gene_content,
                     low_freq_gene_content,
                     core_gene_pairs,
                     master_info) = record_core_core_region(core_genes, gff_name, line, None, previous_core_gene_id,
                                                            previous_core_gene_end_coor, acc_genes_in_region,
                                                            low_freq_genes_in_region, core_gene_pair_distance,
                                                            accessory_gene_content, low_freq_gene_content,
                                                            core_gene_pairs, master_info)

            else:
                # Check if accessory is low frequency - else just regular accessory
                if line[8] in low_freq_genes[gff_name]:
                    low_freq_genes_in_region.append(low_freq_genes[gff_name][line[8]])
                else:
                    # acc_genes_in_region.append(acc_genes[gff_name][line[8]])
                    try:
                        acc_genes_in_region.append(acc_genes[gff_name][line[8]])
                    except KeyError: # TODO - WHAT DOES THIS DO? - Likely search for fragment within composite, as fragments were previously storred in their composit strings.
                        gene_key = [key for key in acc_genes[gff_name].keys() if line[8] in key]
                        if len(gene_key) > 1:
                            acc_genes_in_region.append(acc_genes[gff_name][gene_key][0])

        else:
            # Check if there is a core gene on traversed contig or if a core gene is present on the first contig -
            # if then record it, if not record the accessory and low frequency genes found on contig and reset.
            if previous_core_gene_id != "Sequence_break" and previous_core_gene_id != "":
                if complete_genome:
                    (previous_core_gene_id,
                     previous_core_gene_end_coor,
                     acc_genes_in_region,
                     low_freq_genes_in_region,
                     core_gene_pairs,
                     core_gene_pair_distance,
                     accessory_gene_content,
                     low_freq_gene_content,
                     master_info) = connect_first_n_last_gene_on_contig(core_genes, gff_name, previous_core_gene_id,
                                                                        previous_core_gene_end_coor,
                                                                        first_core_gene_gff_line, acc_genes_in_region,
                                                                        first_core_accessory_content,
                                                                        low_freq_genes_in_region,
                                                                        first_core_low_freq_genes, contig_sizes[line[0]],
                                                                        core_gene_pairs, core_gene_pair_distance,
                                                                        accessory_gene_content, low_freq_gene_content,
                                                                        master_info)

                    first_core_gene = True

                else:
                    # Record the core gene neighbouring a sequence break
                    (previous_core_gene_id,
                     previous_core_gene_end_coor,
                     acc_genes_in_region,
                     low_freq_genes_in_region,
                     core_gene_pair_distance,
                     accessory_gene_content,
                     low_freq_gene_content,
                     core_gene_pairs,
                     master_info) = record_core_core_region(core_genes, gff_name, None, contig_sizes[previous_contig],
                                                            previous_core_gene_id, previous_core_gene_end_coor,
                                                            acc_genes_in_region, low_freq_genes_in_region,
                                                            core_gene_pair_distance, accessory_gene_content,
                                                            low_freq_gene_content, core_gene_pairs, master_info)
            else:
                # Record info on accessory genes on core-less contig, if any accessory genes are present
                coreless_contigs = record_coreless_contig(coreless_contigs, acc_genes_in_region,
                                                          low_freq_genes_in_region, gff_name, previous_contig)

                # Reset accessory and low-frequency gene lists
                acc_genes_in_region = []
                low_freq_genes_in_region = []

                # Set new contig
                previous_contig = line[0]

            # Check if fist gene is core on complete genome, if then record details.
            if line[8] in core_genes[gff_name] and complete_genome:
                # Set information on first core gene to be used when finishing search
                first_core_gene_gff_line = line

                # Set information to be used with next core gene neighbour
                previous_core_gene_end_coor = int(line[4])
                previous_core_gene_id = line[8]

                # Set that first core gene has been observed
                first_core_gene = False

            # Check if first gene on new contig is a core gene, if the record it.
            elif line[8] in core_genes[gff_name]:
                previous_core_gene_id = "Sequence_break"

                # Get the starting position of the first core gene on contig to record the gene.
                # Make it negative to fit the calculation of the distance between genes.
                cur_core_gene_start = -int(line[3])

                (previous_core_gene_id,
                 previous_core_gene_end_coor,
                 acc_genes_in_region,
                 low_freq_genes_in_region,
                 core_gene_pair_distance,
                 accessory_gene_content,
                 low_freq_gene_content,
                 core_gene_pairs,
                 master_info) = record_core_core_region(core_genes, gff_name, line, 0, previous_core_gene_id,
                                                        cur_core_gene_start, acc_genes_in_region,
                                                        low_freq_genes_in_region, core_gene_pair_distance,
                                                        accessory_gene_content, low_freq_gene_content,
                                                        core_gene_pairs, master_info)

            # Add as accessory - if first gene is not core
            else:
                # Check if accessory is low frequency - else just regular accessory
                if line[8] in low_freq_genes[gff_name]:
                    low_freq_genes_in_region.append(low_freq_genes[gff_name][line[8]])
                else:
                    try:
                        acc_genes_in_region.append(acc_genes[gff_name][line[8]])
                    except KeyError:
                        gene_key = [key for key in acc_genes[gff_name].keys() if line[8] in key]
                        if len(gene_key) > 1:
                            acc_genes_in_region.append(acc_genes[gff_name][gene_key][0])

            # Set new contig
            previous_contig = line[0]

    # Check if genome is complete or a single contig. If then add information for last and first core gene, if not
    # then add the first and last core gene as being neighbours to sequence breaks.
    if complete_genome:
        if previous_core_gene_id != "":
            (previous_core_gene_id,
             previous_core_gene_end_coor,
             acc_genes_in_region,
             low_freq_genes_in_region,
             core_gene_pairs,
             core_gene_pair_distance,
             accessory_gene_content,
             low_freq_gene_content,
             master_info) = connect_first_n_last_gene_on_contig(core_genes, gff_name, previous_core_gene_id,
                                                                previous_core_gene_end_coor,
                                                                first_core_gene_gff_line, acc_genes_in_region,
                                                                first_core_accessory_content,
                                                                low_freq_genes_in_region,
                                                                first_core_low_freq_genes, contig_sizes[line[0]],
                                                                core_gene_pairs, core_gene_pair_distance,
                                                                accessory_gene_content, low_freq_gene_content,
                                                                master_info)
        else:
            # Record info on accessory genes on core-less contig, if any accessory genes are present
            record_coreless_contig(coreless_contigs, acc_genes_in_region, low_freq_genes_in_region, gff_name, line[0])

    else:
        # Add first core gene as being neighbour to a sequence break
        (_,
         _,
         _,
         _,
         core_gene_pair_distance,
         accessory_gene_content,
         low_freq_gene_content, core_gene_pairs,
         master_info) = record_core_core_region(core_genes, gff_name, first_core_gene_gff_line, 0,
                                                                         'Sequence_break',
                                                                         9999999999999999,
                                                                         first_core_accessory_content,
                                                                         first_core_low_freq_genes,
                                                                         core_gene_pair_distance,
                                                                         accessory_gene_content,
                                                                         low_freq_gene_content,
                                                                         core_gene_pairs,
                                                                         master_info)
        # Add last core gene as being neighbour to a sequence break
        # if the last core gene has not been recorded already
        if previous_core_gene_id != "Sequence_break":
            (previous_core_gene_id,
             previous_core_gene_end_coor,
             acc_genes_in_region,
             low_freq_genes_in_region,
             core_gene_pair_distance,
             accessory_gene_content,
             low_freq_gene_content, core_gene_pairs,
            master_info) = record_core_core_region(core_genes, gff_name, None, contig_sizes[previous_contig],
                                                                             previous_core_gene_id,
                                                                             previous_core_gene_end_coor,
                                                                             acc_genes_in_region,
                                                                             low_freq_genes_in_region,
                                                                             core_gene_pair_distance,
                                                                             accessory_gene_content,
                                                                             low_freq_gene_content,
                                                                             core_gene_pairs,
                                                                             master_info)
        else:
            # Add a core-less contig if there has been accessory genes:
            coreless_contigs = record_coreless_contig(coreless_contigs, acc_genes_in_region, low_freq_genes_in_region, gff_name, line[0])

    return core_gene_pairs, core_gene_pair_distance, accessory_gene_content, \
           low_freq_gene_content, master_info, coreless_contigs


def segment_genome_content(input_gff_file, core_genes, low_freq_genes, acc_gene_dict, complete_genomes, source_program,
                           annotate, gene_data_dict, corrected_dir, tmp_folder_path, discard_corrected, logger):
    """
    Single function segmenting the gff into core gene regions to be used for simple multi processing
    :param input_gff_file: File-path to the given gff file to be segmented
    :param core_genes: Dictionary over core genes
    :param low_freq_genes: Dictionary over low-frequency genes
    :param acc_gene_dict: Dictionary over accessory genes
    :param complete_genomes: Bool indicating if this genome should be considered as a complete genome
    :param source_program: String indicating if program comes from Roary or Panaroo.
    :param annotate: Bool to indicate if refound genes should be annotated
    :param gene_data_dict: Dict of genes, annotations, names, and sequences found in the gene_data.csv file from Panaroo
    :param corrected_dir: File path to directory where corrected Gff files are to be stored.
    :param tmp_folder_path: Path to the temporary working folder.
    :param discard_corrected: Bool indicating if corrected Gff files should be preserved as an output
    :param logger: Progran logger

    :return input_gff_file: File path to the gff being searched
    :return core_genes: Dict of core genes passed to genomes and the pan-genome clusters.
    :return low_freq_genes: Same structure as core_genes, but for low-frequency genes.
    :return acc_gene_dict: Same structure as core_genes, but for accessory genes.
    :return i: The index of the gff in the larger scheme of the analysis
    :return complete_genomes: List of genomes given as complete by the user.
    """

    # Correct input gff file
    # Add in the refound genes into the gff files and print the corrected GFF files.
    if source_program == "Panaroo" and annotate:
        # check if not already corrected file and if any gene is to be inserted at all
        if "_corrected" not in input_gff_file and any([x in input_gff_file for x in list(gene_data_dict)]):
            input_gff_file = annotate_refound_genes(input_gff_file, gene_data_dict, tmp_folder_path, corrected_dir, logger)

    gff_generator = parse_gff(input_gff_file)
    return_data = segment_gff_content(gff_generator=gff_generator,
                                      gff_path=input_gff_file,
                                      core_genes=core_genes,
                                      low_freq_genes=low_freq_genes,
                                      acc_genes=acc_gene_dict,
                                      complete_genomes=complete_genomes)

    if "_corrected" in input_gff_file and discard_corrected:
        os.remove(input_gff_file)

    return return_data



import os
import csv
from math import ceil, floor
import gffutils


def add_gene_to_dict(main_dict, gene, pan_gene_name, genome):
    """
    Function to add a gene to a given dictionary
    :param main_dict: Dict of genes from genomes. A dict of dicts, with first set of keys being genomes, second is locus_tags with pan-genome gene being the key.
    :param gene: The gene in question from a specific genome (locus_tag)
    :param pan_gene_name: The name of the pan-genome gene (cluster) to which the above gene belongs.
    :param genome: The name of the genome in question
    :return: returns the dict to be used further
    """
    if ';' in gene:
        for gene_part in gene.split(';'):  # TODO - NOTE! HERE BOTH GENES IN A PAIR IS ADDED as separate key/value-pairs
            main_dict[genome][gene_part] = pan_gene_name
    else:
        main_dict[genome][gene] = pan_gene_name

    return main_dict


def check_fragmented_gene(fragment_info, input_gffs, tmp_folder_path):
    """
    Function that check for that placement of fragmented gene parts, to determine if they are neighbouring or have some genomic feature between them
    :param fragment_info: List of genes that are found to be fragmented, one composite of fragments for each index
    :param input_gffs: A list of file-paths to the gff files given as input
    :param tmp_folder_path: A file-path to the temporary folder of the Corekaburra run
    :return: A List of booleans indicating if a fragments has nothing in between fragments (True) or not (False)
    """
    return_list = []
    for fragment in fragment_info:
        # split the two fragments
        fragment_pieces = fragment[0].split(';')

        # Get the name of the genome
        genome = fragment[1]

        # Get the gff and its path
        try:
            gff_file = [file for file in input_gffs if f'{genome}.gff' in file][0]
        except IndexError:
            raise NotImplementedError(f'No gff match was found when searching fragments for genome: {genome}')

        # Construct gff database to be searched
        db_name = os.path.join(tmp_folder_path, f'{genome}_db')
        if not os.path.isfile(db_name):
            gffutils.create_db(gff_file, db_name, force_gff=True)

        # Attach database
        gff_database = gffutils.FeatureDB(db_name)

        # Check that all fragments are on the same contig.
        first_fragment_contig = gff_database[fragment_pieces[0]][0]
        frag_same_contig = all([first_fragment_contig == gff_database[fragment][0] for fragment in fragment_pieces])
        if frag_same_contig:
            # Get all coordinates
            frag_coors = []
            for frag in fragment_pieces:
                frag_coors.append(gff_database[frag][3])
                frag_coors.append(gff_database[frag][4])

            # Construct region to be searched for annotations between fragments:
            max_frag_coor = max(frag_coors)
            min_frag_coor = min(frag_coors)
            region = (first_fragment_contig, min_frag_coor, max_frag_coor)

            # Find all features that are completely within the region
            region_features = gff_database.region(region=region, completely_within=True)

            # find all genes that are not part of the fragmented gene
            region_locus_tags = set([feature[8]['locus_tag'][0] for feature in region_features])
            excess_genes = region_locus_tags.difference(fragment_pieces)

            # check the number of excess genes, if any then False to being core
            if len(excess_genes) > 0:
                return_list.append(False)
            else:
                return_list.append(True)
        else:
            return_list.append(False)

    return return_list
    # TODO - find out what the non-closed file problem is here! Can be seen when running unit-tests.


def read_gene_presence_absence(pres_abs_file, core_gene_presence, low_freq_gene, source_program, input_gffs, tmp_folder_path, logger):
    """
    Function that pass a Roary style gene presence/absence file.
    :param pres_abs_file: File path to the gene presence/absence file identified
    :param core_gene_presence: The ratio of genomes in which a gene must present, to be seen as a core gene
    :param low_freq_gene: The ratio of genomes in which a gene must not surpass, to be seen as a low-frequency gene
    :param source_program: The program from which the pan-genome was produced
    :param input_gffs: A list of file-paths to the gff files given as input
    :param tmp_folder_path: A file-path to the temporary folder of the Corekaburra run
    :param logger: Program logger
    :return: Directories of directories of core and low frequency genes, and a directory of pan genome clusters and their annotation.
    """

    # Open the presence/absense file to index gene into core, accessory, or low-frequency genes
    with open(pres_abs_file, 'r', newline='', ) as gene_presence_absence:
        # Read column header line
        gff_file_names = gene_presence_absence.readline()
        # Strip for whitespace
        gff_file_names = gff_file_names.strip()
        # split column names
        gff_file_names = gff_file_names.split(',')

        # Remove the quotes from Rorary input
        if source_program == 'Roary':
            gff_file_names = [filename.replace('"', '') for filename in gff_file_names]

        # Index gff filenames and column position in dict for better search
        gff_file_dict = {}
        for i, gff_name in enumerate(gff_file_names[14:]):
            gff_file_dict[gff_name] = i

        # Read remaining lines and construct a nested dicts one dict for each genome and its core genes,
        # and a dict for low frequency genes found in less than set percent of isolates

        # Initialise reader object to read remaining lines
        reader = csv.reader(gene_presence_absence, delimiter=',')
        # Counters
        core_gene_number = 0
        low_freq_gene_number = 0
        acc_gene_number = 0

        # Determine number of isolates that represent core and low frequency genes
        core_gene_isolate_presence = floor(len(gff_file_dict.keys()) * core_gene_presence)
        low_freq_gene_isolate_presence = ceil(len(gff_file_dict.keys()) * low_freq_gene)

        logger.info(f"------------Opening the gene presence/absence file------------\n"
                    f"Core genes must be found in {core_gene_isolate_presence} or more genomes\n"
                    f"Low frequency genes must be found in less than {low_freq_gene_isolate_presence} genomes\n")

        # initialise dict of dicts to hold genes from each gffs and to be returned
        core_gene_dict = {item: {} for item in gff_file_names[14:]}
        low_freq_gene_dict = {item: {} for item in gff_file_names[14:]}
        acc_gene_dict = {item: {} for item in gff_file_names[14:]}

        # Read lines from file and determine if core, low frequency or 'regular' accessory and record annotations
        for line in reader:
            # Remove quotes if Roary
            if source_program == 'Roary':
                line = [element.replace('"', '') for element in line]

            # Get number of genes in line and average presence of genes in genomes
            gene_isolate_presence = int(line[3])
            no_seq_presence = int(line[4])

            # Check if core gene, if then add annotations to genomes
            # Check if gene is present in all genomes and no one gene is fragmented
            if core_gene_isolate_presence <= gene_isolate_presence == no_seq_presence:
                # Add gene cluster to genomes
                for genome in core_gene_dict.keys():
                    # Check if there is an annotation for the given genome
                    if len(line[14 + gff_file_dict[genome]]) > 0:
                        core_gene_dict[genome][line[14+gff_file_dict[genome]]] = line[0]
                core_gene_number += 1

            # Check if gene is present in all genomes, but more than one copy is present
            elif core_gene_isolate_presence <= gene_isolate_presence:
                # Identify annotations for genomes that are fragmented genes
                fragment_info = [[genes, gff] for genes, gff in zip(line[14:], gff_file_names[14:]) if ';' in genes]

                # Check that each annotation is neighboring the other annotation.
                return_list = check_fragmented_gene(fragment_info, input_gffs, tmp_folder_path) # TODO - If a core gene is found to be made up of fragments not places close enough (With something in between) should this then not be subtracted from the core gene count? - How would this be handled if there is a gff that is not given as input?

                # Check if gene was found to be a core gene
                if all(return_list):
                    # Add the gene to the annotation dict
                    for genome in core_gene_dict:
                        # Get the annoations for a specific genome
                        genes_in_genome = line[14 + gff_file_dict[genome]]
                        # If there is an annotation add id
                        if len(genes_in_genome) > 0:
                            # Check if genome has fragments of genes,
                            # if then add them all to the annotation dict,
                            # if not then just ad the single annotation
                            add_gene_to_dict(core_gene_dict, genes_in_genome, line[0], genome)
                    core_gene_number += 1

                else:
                    # Check if low frequency, if then add else then add as normal accessory
                    if low_freq_gene_isolate_presence >= gene_isolate_presence == no_seq_presence:
                        for genome in low_freq_gene_dict.keys():
                            if len(line[14 + gff_file_dict[genome]]) > 0:
                                add_gene_to_dict(low_freq_gene_dict, line[14 + gff_file_dict[genome]], line[0], genome)
                        low_freq_gene_number += 1
                    else:
                        for genome in acc_gene_dict.keys():
                            if len(line[14 + gff_file_dict[genome]]) > 0:
                                add_gene_to_dict(acc_gene_dict, line[14 + gff_file_dict[genome]], line[0], genome)
                        acc_gene_number += 1

            # Check if accessory if then add annotation to genomes
            elif low_freq_gene_isolate_presence >= gene_isolate_presence == no_seq_presence:
                for genome in low_freq_gene_dict.keys():
                    if len(line[14+gff_file_dict[genome]]) > 0:
                        add_gene_to_dict(low_freq_gene_dict, line[14 + gff_file_dict[genome]], line[0], genome)
                low_freq_gene_number += 1

            # If not core or low frequency count as regular accessory
            else:
                for genome in acc_gene_dict.keys():
                    if len(line[14+gff_file_dict[genome]]) > 0:
                        add_gene_to_dict(acc_gene_dict, line[14 + gff_file_dict[genome]], line[0], genome)
                acc_gene_number += 1

    logger.info("A total of:\n"
                f"{core_gene_number} core gene clusters were identified\n"
                f"{low_freq_gene_number} low frequency gene clusters were identified\n"
                f"{acc_gene_number} intermediate accessory gene clusters were identified\n")

    # Remove gff databases
    files_in_tmp = os.listdir(tmp_folder_path)
    gff_dbs = [file for file in files_in_tmp if '_db' in file]
    [os.remove(os.path.join(tmp_folder_path, db)) for db in gff_dbs]

    return core_gene_dict, low_freq_gene_dict, acc_gene_dict


if __name__ == '__main__':
    pass
